<?php

namespace Drupal\a12s_maps_sync\Plugin\SourceHandler;

use Drupal\a12s_maps_sync\Entity\ConverterInterface;
use Drupal\a12s_maps_sync\Maps\Attribute;
use Drupal\a12s_maps_sync\Maps\Link;
use Drupal\a12s_maps_sync\Plugin\SourceHandlerPluginBase;
use Drupal\Core\Entity\EntityInterface;

/**
 * Plugin implementation of the a12s_maps_sync_source_handler.
 *
 * @SourceHandler(
 *   id = "object",
 *   label = @Translation("Object"),
 *   description = @Translation("Object.")
 * )
 */
class MapsObject extends SourceHandlerPluginBase {

  use AttributesTrait;

  /**
   * {@inheritdoc}
   * @param \Drupal\a12s_maps_sync\Entity\ConverterInterface $converter
   */
  public function getFilterTypes(ConverterInterface $converter): array {
    $filterTypes = [
      'object_nature' => $this->t('Object nature'),
      'object_type' => $this->t('Object type'),
      'class' => $this->t('Object class'),
      'parent_id' => $this->t('Parent ID'),
    ];

    /** @var \Drupal\a12s_maps_sync\MapsApi $mapsApi */
    $mapsApi = \Drupal::service('a12s_maps_sync.maps_api');

    // Get the available object properties.
    $properties = $mapsApi->getObjectProperties($converter->getProfile(), ['type' => 'attribute']);
    foreach ($properties as $property) {
      $filterTypes['Properties'][$property] = $property;
    }
    if (!empty($filterTypes['Properties'])) {
      asort($filterTypes['Properties']);
    }

    $filterTypes['Attributes'] = $this->getAttributesFilters($converter);

    return $filterTypes;
  }

  /**
   * {@inheritdoc}
   */
  public function getCountData(ConverterInterface $converter, array $filters, int $fromTime): int {
    // @todo use dependency injection.
    /** @var \Drupal\a12s_maps_sync\MapsApi $mapsApi */
    $mapsApi = \Drupal::service('a12s_maps_sync.maps_api');
    return $mapsApi->getObjects($converter->getProfile(), $filters, -1, $fromTime, TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function getData(ConverterInterface $converter, array $filters, int $limit, int $fromTime = 0): array {
    /** @var \Drupal\a12s_maps_sync\MapsApi $mapsApi */
    $mapsApi = \Drupal::service('a12s_maps_sync.maps_api');
    $data = $mapsApi->getObjects($converter->getProfile(), $filters, $limit, $fromTime);

    $return = [];
    foreach ($data as $datum) {
      $object = new \Drupal\a12s_maps_sync\Maps\MapsObject(
        $datum['id'],
        $datum['parent_id'],
        $datum['source_id'],
        $datum['code'],
        $datum['nature'],
        $datum['type'],
        $datum['classes'],
        $datum['status'],
        $datum['sequence'],
        (int) $datum['imported'],
        $datum['config_type'],
        $datum['medias'],
      );

      $object->setProfile($converter->getProfile()->id());
      $object->setConverter($converter->id());

      // Add the attributes.
      $attributes = $this->generateAttributes($converter->getProfile(), $datum);
      $object->setAttributes($attributes);

      // Add the medias.
      $medias = [];
      foreach ($datum['medias'] as $mediaData) {
        $medias[] = new \Drupal\a12s_maps_sync\Maps\Media(
          $mediaData['id'],
          $mediaData['extension'],
          $mediaData['url'],
          $mediaData['filename'],
          $mediaData['type'],
          $mediaData['weight'],
          $mediaData['imported'],
        );
      }
      $object->setMedias($medias);

      // Add the links.
      $links = [];
      foreach ($datum['links'] as $linkData) {
        $links[] = new Link(
          $linkData['source_id'],
          $linkData['target_id'],
          $linkData['type_id'],
          $linkData['count'],
        );
      }
      $object->setLinks($links);

      $object->setProperties($datum['properties']);

      $return[] = $object;
    }

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function getMappingSources(): array {
    return [
      'Properties' => [
        'id' => $this->t('ID'),
        'source_id' => $this->t('Origin ID'),
        'parent_id' => $this->t('Parent ID'),
        'code' => $this->t('Code'),
        'medias' => $this->t('Medias'),
        'links' => $this->t('Links'),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getAllowedGidEntityKeys(): array {
    return ['id', 'source_id', 'code'];
  }

}
